package com.fox2code.mmm.repo;

import com.fox2code.mmm.MainApplication;
import com.fox2code.mmm.utils.io.Http;
import com.fox2code.mmm.utils.realm.ModuleListCache;
import com.fox2code.mmm.utils.realm.ReposList;

import org.json.JSONArray;
import org.json.JSONObject;

import java.io.File;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;

import io.realm.Realm;
import io.realm.RealmConfiguration;
import io.realm.RealmResults;
import timber.log.Timber;

public class RepoUpdater {
    public final RepoData repoData;
    public byte[] indexRaw;
    private List<RepoModule> toUpdate;
    private Collection<RepoModule> toApply;

    public RepoUpdater(RepoData repoData) {
        this.repoData = repoData;
    }

    public int fetchIndex() {
        if (!RepoManager.getINSTANCE().hasConnectivity()) {
            this.indexRaw = null;
            this.toUpdate = Collections.emptyList();
            this.toApply = Collections.emptySet();
            return 0;
        }
        if (!this.repoData.isEnabled()) {
            this.indexRaw = null;
            this.toUpdate = Collections.emptyList();
            this.toApply = Collections.emptySet();
            return 0;
        }
        // if we shouldn't update, get the values from the ModuleListCache realm
        if (!this.repoData.shouldUpdate()) {
            RealmConfiguration realmConfiguration = new RealmConfiguration.Builder()
                    .name("ModuleListCache.realm")
                    .schemaVersion(1)
                    .modules(new ModuleListCache())
                    .build();
            Realm realm = Realm.getInstance(realmConfiguration);
            RealmResults<ModuleListCache> results = realm.where(ModuleListCache.class).equalTo("repoId", this.repoData.id).findAll();
            // reposlist realm
            RealmConfiguration realmConfiguration2 = new RealmConfiguration.Builder().name("ReposList.realm").allowQueriesOnUiThread(true).allowWritesOnUiThread(true).directory(MainApplication.getINSTANCE().getDataDirWithPath("realms")).schemaVersion(1).build();
            Realm realm2 = Realm.getInstance(realmConfiguration2);
            ReposList reposList = realm2.where(ReposList.class).equalTo("id", this.repoData.id).findFirst();
            this.toUpdate = Collections.emptyList();
            this.toApply = new HashSet<>();
            for (ModuleListCache moduleListCache : results) {
                RepoData repoData = RepoManager.getINSTANCE().get(Objects.requireNonNull(reposList).getUrl());
                this.toApply.add(new RepoModule(repoData, moduleListCache.getId()));
            }
            this.toApply = new HashSet<>(this.toUpdate);
            return this.toUpdate.size();
        }
        try {
            if (!this.repoData.prepare()) {
                this.indexRaw = null;
                this.toUpdate = Collections.emptyList();
                this.toApply = this.repoData.moduleHashMap.values();
                return 0;
            }
            this.indexRaw = Http.doHttpGet(this.repoData.getUrl(), false);
            this.toUpdate = this.repoData.populate(new JSONObject(new String(this.indexRaw, StandardCharsets.UTF_8)));
            // Since we reuse instances this should work
            this.toApply = new HashSet<>(this.repoData.moduleHashMap.values());
            this.toApply.removeAll(this.toUpdate);
            // Return repo to update
            return this.toUpdate.size();
        } catch (
                Exception e) {
            Timber.e(e);
            this.indexRaw = null;
            this.toUpdate = Collections.emptyList();
            this.toApply = Collections.emptySet();
            return 0;
        }
    }

    public List<RepoModule> toUpdate() {
        return this.toUpdate;
    }

    public Collection<RepoModule> toApply() {
        return this.toApply;
    }

    public boolean finish() {
        boolean success = this.indexRaw != null;
        // If repo is not enabled we don't need to do anything, just return true
        if (!this.repoData.isEnabled()) {
            return true;
        }
        if (this.indexRaw != null) {
            try {
                // iterate over modules, using this.supportedProperties as a template to attempt to get each property from the module. everything that is not null is added to the module
                // use realm to insert to
                // props avail:
                File cacheRoot = MainApplication.getINSTANCE().getDataDirWithPath("realms/repos" + this.repoData.id);
                RealmConfiguration realmConfiguration = new RealmConfiguration.Builder().name("ModuleListCache.realm").schemaVersion(1).deleteRealmIfMigrationNeeded().allowWritesOnUiThread(true).allowQueriesOnUiThread(true).directory(cacheRoot).build();
                // array with module info default values
                // supported properties for a module
                //id=<string>
                //name=<string>
                //version=<string>
                //versionCode=<int>
                //author=<string>
                //description=<string>
                //minApi=<int>
                //maxApi=<int>
                //minMagisk=<int>
                //needRamdisk=<boolean>
                //support=<url>
                //donate=<url>
                //config=<package>
                //changeBoot=<boolean>
                //mmtReborn=<boolean>
                // extra properties only useful for the database
                //repoId=<string>
                //installed=<boolean>
                //installedVersionCode=<int> (only if installed)
                //
                // all except first six can be null
                // this.indexRaw is the raw index file (json)
                JSONObject modules = new JSONObject(new String(this.indexRaw, StandardCharsets.UTF_8));
                JSONArray modulesArray;
                // androidacy repo uses "data" key, others should use "modules" key. Both are JSONArrays
                if (this.repoData.getName().equals("Androidacy Modules Repo")) {
                    // get modules from "data" key. This is a JSONArray so we need to convert it to a JSONObject
                    modulesArray = modules.getJSONArray("data");
                } else {
                    // get modules from "modules" key. This is a JSONArray so we need to convert it to a JSONObject
                    modulesArray = modules.getJSONArray("modules");
                }
                Realm realm = Realm.getInstance(realmConfiguration);
                // drop old data
                realm.beginTransaction();
                realm.where(ModuleListCache.class).equalTo("repoId", this.repoData.id).findAll().deleteAllFromRealm();
                realm.commitTransaction();
                // iterate over modules. pls dont hate me for this, its ugly but it works
                for (int n = 0; n < modulesArray.length(); n++) {
                    // get module
                    JSONObject module = modulesArray.getJSONObject(n);
                    try {
                        // get module id
                        // if codename is present, prefer that over id
                        String id;
                        if (module.has("codename") && !module.getString("codename").equals("")) {
                            id = module.getString("codename");
                        } else {
                            id = module.getString("id");
                        }
                        // get module name
                        String name = module.getString("name");
                        // get module version
                        String version = module.getString("version");
                        // get module version code
                        int versionCode = module.getInt("versionCode");
                        // get module author
                        String author = module.getString("author");
                        // get module description
                        String description = module.getString("description");
                        // get module min api
                        String minApi;
                        if (module.has("minApi") && !module.getString("minApi").equals("")) {
                            minApi = module.getString("minApi");
                        } else {
                            minApi = "0";
                        }
                        // coerce min api to int
                        int minApiInt = Integer.parseInt(minApi);
                        // get module max api and set to 0 if it's "" or null
                        String maxApi;
                        if (module.has("maxApi") && !module.getString("maxApi").equals("")) {
                            maxApi = module.getString("maxApi");
                        } else {
                            maxApi = "0";
                        }
                        // coerce max api to int
                        int maxApiInt = Integer.parseInt(maxApi);
                        // get module min magisk
                        String minMagisk;
                        if (module.has("minMagisk") && !module.getString("minMagisk").equals("")) {
                            minMagisk = module.getString("minMagisk");
                        } else {
                            minMagisk = "0";
                        }
                        // coerce min magisk to int
                        int minMagiskInt = Integer.parseInt(minMagisk);
                        // get module need ramdisk
                        boolean needRamdisk;
                        if (module.has("needRamdisk")) {
                            needRamdisk = module.getBoolean("needRamdisk");
                        } else {
                            needRamdisk = false;
                        }
                        // get module support
                        String support;
                        if (module.has("support")) {
                            support = module.getString("support");
                        } else {
                            support = "";
                        }
                        // get module donate
                        String donate;
                        if (module.has("donate")) {
                            donate = module.getString("donate");
                        } else {
                            donate = "";
                        }
                        // get module config
                        String config;
                        if (module.has("config")) {
                            config = module.getString("config");
                        } else {
                            config = "";
                        }
                        // get module change boot
                        boolean changeBoot;
                        if (module.has("changeBoot")) {
                            changeBoot = module.getBoolean("changeBoot");
                        } else {
                            changeBoot = false;
                        }
                        // get module mmt reborn
                        boolean mmtReborn;
                        if (module.has("mmtReborn")) {
                            mmtReborn = module.getBoolean("mmtReborn");
                        } else {
                            mmtReborn = false;
                        }
                        // get module repo id
                        String repoId = this.repoData.id;
                        // get module installed
                        boolean installed = false;
                        // get module installed version code
                        int installedVersionCode = 0;
                        // insert module to realm
                        // first create a collection of all the properties
                        // then insert to realm
                        // then commit
                        // then close
                        if (realm.isInTransaction()) {
                            realm.cancelTransaction();
                        }
                        // create a realm object and insert or update it
                        // add everything to the realm object
                        realm.beginTransaction();
                        ModuleListCache moduleListCache = realm.createObject(ModuleListCache.class, id);
                        moduleListCache.setName(name);
                        moduleListCache.setVersion(version);
                        moduleListCache.setVersionCode(versionCode);
                        moduleListCache.setAuthor(author);
                        moduleListCache.setDescription(description);
                        moduleListCache.setMinApi(minApiInt);
                        moduleListCache.setMaxApi(maxApiInt);
                        moduleListCache.setMinMagisk(minMagiskInt);
                        moduleListCache.setNeedRamdisk(needRamdisk);
                        moduleListCache.setSupport(support);
                        moduleListCache.setDonate(donate);
                        moduleListCache.setConfig(config);
                        moduleListCache.setChangeBoot(changeBoot);
                        moduleListCache.setMmtReborn(mmtReborn);
                        moduleListCache.setRepoId(repoId);
                        moduleListCache.setInstalled(installed);
                        moduleListCache.setInstalledVersionCode(installedVersionCode);
                        realm.copyToRealmOrUpdate(moduleListCache);
                        realm.commitTransaction();
                    } catch (
                            Exception e) {
                        e.printStackTrace();
                        Timber.w("Failed to get module info from module " + module + " in repo " + this.repoData.id + " with error " + e.getMessage());
                    }
                }
                realm.close();
            } catch (
                    Exception e) {
                e.printStackTrace();
            }
            this.indexRaw = null;
            RealmConfiguration realmConfiguration2 = new RealmConfiguration.Builder().name("ReposList.realm").allowQueriesOnUiThread(true).allowWritesOnUiThread(true).directory(MainApplication.getINSTANCE().getDataDirWithPath("realms")).schemaVersion(1).build();
            Realm realm2 = Realm.getInstance(realmConfiguration2);
            if (realm2.isInTransaction()) {
                realm2.cancelTransaction();
            }
            // set lastUpdate
            realm2.executeTransaction(r -> {
                ReposList repoListCache = r.where(ReposList.class).equalTo("id", this.repoData.id).findFirst();
                if (repoListCache != null) {
                    repoListCache.setLastUpdate((int) System.currentTimeMillis());
                }
            });
        }
        this.toUpdate = null;
        this.toApply = null;
        return success;
    }

}
